# Copyright 2023 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from gpjax.gps import Prior
from gpjax.kernels.stationary.rbf import RBF
from gpjax.likelihoods import Gaussian
from jax import config

config.update("jax_enable_x64", True)

from jax_decision_making.test_functions import (
    AbstractContinuousTestFunction,
    Forrester,
    LogarithmicGoldsteinPrice,
)
from jax_decision_making.utils import (
    OBJECTIVE,
    build_function_evaluator,
    get_best_latent_observation_val,
)
from gpjax.typing import (
    Array,
    Float,
    KeyArray,
)
import jax.numpy as jnp
import jax.random as jr
import pytest


def test_build_function_evaluator():
    def _square(x: Float[Array, "N 1"]) -> Float[Array, "N 1"]:
        return x**2

    def _cube(x: Float[Array, "N 1"]) -> Float[Array, "N 1"]:
        return x**3

    functions = {OBJECTIVE: _square, "CONSTRAINT": _cube}
    fn_evaluator = build_function_evaluator(functions)
    x = jnp.array([[2.0, 3.0]])
    datasets = fn_evaluator(x)
    assert datasets.keys() == functions.keys()
    assert jnp.equal(datasets[OBJECTIVE].X, x).all()
    assert jnp.equal(datasets[OBJECTIVE].y, _square(x)).all()
    assert jnp.equal(datasets["CONSTRAINT"].X, x).all()
    assert jnp.equal(datasets["CONSTRAINT"].y, _cube(x)).all()


@pytest.mark.parametrize(
    "test_target_function",
    [(Forrester()), (LogarithmicGoldsteinPrice())],
)
@pytest.mark.parametrize("key", [jr.PRNGKey(42), jr.PRNGKey(10)])
def test_get_best_observation(
    test_target_function: AbstractContinuousTestFunction,
    key: KeyArray,
):
    obs_stddev = 0.5
    dataset = test_target_function.generate_dataset(
        num_points=100, key=key, obs_stddev=obs_stddev
    )
    mean_fn = test_target_function
    kernel = RBF(lengthscale=1.0 * jnp.ones(dataset.X.shape[1]))
    prior = Prior(kernel=kernel, mean_function=mean_fn)
    likelihood = Gaussian(num_datapoints=dataset.n, obs_stddev=obs_stddev)
    posterior = prior * likelihood
    expected_best_obs = jnp.min(posterior(dataset.X, dataset).mean())
    actual_best_obs = get_best_latent_observation_val(
        posterior=posterior, dataset=dataset
    )
    assert jnp.equal(expected_best_obs, actual_best_obs)



# Copyright 2023 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from jax import config

config.update("jax_enable_x64", True)

from jax_decision_making.test_functions import PoissonTestFunction
from gpjax.typing import KeyArray
import jax.numpy as jnp
import jax.random as jr
import pytest


@pytest.mark.parametrize("test_function", [PoissonTestFunction()])
@pytest.mark.filterwarnings("ignore:y is not of type float64")
def test_correct_dtypes(test_function: PoissonTestFunction):
    dataset = test_function.generate_dataset(10, jr.key(42))
    test_x = test_function.generate_test_points(10, jr.key(42))
    assert dataset.X.dtype == jnp.float64
    assert dataset.y.dtype == jnp.integer
    assert test_x.dtype == jnp.float64


@pytest.mark.parametrize(
    "test_function, dimensionality",
    [(PoissonTestFunction(), 1)],
)
@pytest.mark.parametrize("num_samples", [1, 10, 100])
def test_test_points_shape(
    test_function: PoissonTestFunction, dimensionality: int, num_samples: int
):
    test_X = test_function.generate_test_points(num_samples, jr.key(42))
    assert test_X.shape == (num_samples, dimensionality)


@pytest.mark.parametrize(
    "test_function, dimensionality",
    [(PoissonTestFunction(), 1)],
)
@pytest.mark.parametrize("num_samples", [1, 10, 100])
@pytest.mark.filterwarnings("ignore:y is not of type float64")
def test_dataset_shapes(
    test_function: PoissonTestFunction, dimensionality: int, num_samples: int
):
    dataset = test_function.generate_dataset(num_samples, jr.key(42))
    assert dataset.X.shape == (num_samples, dimensionality)
    assert dataset.y.shape == (num_samples, 1)


@pytest.mark.parametrize("test_function", [PoissonTestFunction()])
@pytest.mark.parametrize("num_samples", [1, 10, 100])
@pytest.mark.parametrize("key", [jr.key(42), jr.key(10)])
@pytest.mark.filterwarnings("ignore:y is not of type float64")
def test_same_key_same_dataset(
    test_function: PoissonTestFunction, num_samples: int, key: KeyArray
):
    dataset_one = test_function.generate_dataset(num_samples, key)
    dataset_two = test_function.generate_dataset(num_samples, key)
    assert jnp.equal(dataset_one.X, dataset_two.X).all()
    assert jnp.equal(dataset_one.y, dataset_two.y).all()


@pytest.mark.parametrize("test_function", [PoissonTestFunction()])
@pytest.mark.parametrize("num_samples", [10, 100])
@pytest.mark.parametrize("key", [jr.key(42), jr.key(10)])
@pytest.mark.filterwarnings("ignore:y is not of type float64")
def test_different_key_different_dataset(
    test_function: PoissonTestFunction, num_samples: int, key: KeyArray
):
    dataset_one = test_function.generate_dataset(num_samples, key)
    key, _ = jr.split(key)
    dataset_two = test_function.generate_dataset(num_samples, key)
    assert not jnp.equal(dataset_one.X, dataset_two.X).all()
    assert not jnp.equal(dataset_one.y, dataset_two.y).all()


@pytest.mark.parametrize("test_function", [PoissonTestFunction()])
@pytest.mark.parametrize("num_samples", [1, 10, 100])
@pytest.mark.parametrize("key", [jr.key(42), jr.key(10)])
@pytest.mark.filterwarnings("ignore:y is not of type float64")
def test_same_key_same_test_points(
    test_function: PoissonTestFunction, num_samples: int, key: KeyArray
):
    test_points_one = test_function.generate_test_points(num_samples, key)
    test_points_two = test_function.generate_test_points(num_samples, key)
    assert jnp.equal(test_points_one, test_points_two).all()


@pytest.mark.parametrize("test_function", [PoissonTestFunction()])
@pytest.mark.parametrize("num_samples", [1, 10, 100])
@pytest.mark.parametrize("key", [jr.key(42), jr.key(10)])
@pytest.mark.filterwarnings("ignore:y is not of type float64")
def test_different_key_different_test_points(
    test_function: PoissonTestFunction, num_samples: int, key: KeyArray
):
    test_points_one = test_function.generate_test_points(num_samples, key)
    key, _ = jr.split(key)
    test_points_two = test_function.generate_test_points(num_samples, key)
    assert not jnp.equal(test_points_one, test_points_two).all()



# Copyright 2023 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from beartype.typing import (
    Callable,
    Dict,
    Final,
)
from gpjax.dataset import Dataset
from gpjax.gps import AbstractPosterior
from gpjax.typing import (
    Array,
    Float,
)
import jax.numpy as jnp

OBJECTIVE: Final[str] = "OBJECTIVE"
"""
Tag for the objective dataset/function in standard utility functions.
"""


FunctionEvaluator = Callable[[Float[Array, "N D"]], Dict[str, Dataset]]
"""
Type alias for function evaluators, which take an array of points of shape $[N, D]$
and evaluate a set of functions at each point, returning a mapping from function tags
to datasets of the evaluated points. This is the same as the `Observer` in Trieste:
https://github.com/secondmind-labs/trieste/blob/develop/trieste/observer.py
"""


def build_function_evaluator(
    functions: Dict[str, Callable[[Float[Array, "N D"]], Float[Array, "N 1"]]],
) -> FunctionEvaluator:
    """
    Takes a dictionary of functions and returns a `FunctionEvaluator` which can be
    used to evaluate each of the functions at a supplied set of points and return a
    dictionary of datasets storing the evaluated points.
    """
    return lambda x: {tag: Dataset(x, f(x)) for tag, f in functions.items()}


def get_best_latent_observation_val(
    posterior: AbstractPosterior, dataset: Dataset
) -> Float[Array, ""]:
    """
    Takes a posterior and dataset and returns the best (latent) function value in the
    dataset, corresponding to the minimum of the posterior mean value evaluated at
    locations in the dataset. In the noiseless case, this corresponds to the minimum
    value in the dataset.
    """
    return jnp.min(posterior(dataset.X, dataset).mean())



# Copyright 2023 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from jax import config

config.update("jax_enable_x64", True)

from beartype.typing import Callable
from jax_decision_making.test_functions.continuous_functions import (
    AbstractContinuousTestFunction,
    Forrester,
    LogarithmicGoldsteinPrice,
)
from jax_decision_making.utility_functions.thompson_sampling import ThompsonSampling
from jax_decision_making.utils import OBJECTIVE
from gpjax.typing import KeyArray
import jax.random as jr
import pytest

from tests.utils import generate_dummy_conjugate_posterior


@pytest.mark.parametrize("num_rff_features", [0, -1, -10])
@pytest.mark.filterwarnings(
    "ignore::UserWarning"
)  # Sampling with tfp causes JAX to raise a UserWarning due to some internal logic around jnp.argsort
def test_thompson_sampling_invalid_rff_num_raises_error(num_rff_features: int):
    key = jr.key(42)
    forrester = Forrester()
    dataset = forrester.generate_dataset(num_points=10, key=key)
    posterior = generate_dummy_conjugate_posterior(dataset)
    posteriors = {OBJECTIVE: posterior}
    datasets = {OBJECTIVE: dataset}
    with pytest.raises(ValueError):
        ts_utility_builder = ThompsonSampling(num_features=num_rff_features)
        ts_utility_builder.build_utility_function(
            posteriors=posteriors, datasets=datasets, key=key
        )


@pytest.mark.parametrize(
    "test_target_function",
    [(Forrester()), (LogarithmicGoldsteinPrice())],
)
@pytest.mark.parametrize("num_test_points", [50, 100])
@pytest.mark.parametrize("key", [jr.key(42), jr.key(10)])
@pytest.mark.filterwarnings(
    "ignore::UserWarning"
)  # Sampling with tfp causes JAX to raise a UserWarning due to some internal logic around jnp.argsort
def test_thompson_sampling_utility_function_same_key_same_function(
    test_target_function: AbstractContinuousTestFunction,
    num_test_points: int,
    key: KeyArray,
):
    dataset = test_target_function.generate_dataset(num_points=10, key=key)
    posterior = generate_dummy_conjugate_posterior(dataset)
    posteriors = {OBJECTIVE: posterior}
    datasets = {OBJECTIVE: dataset}
    ts_utility_builder_one = ThompsonSampling(num_features=100)
    ts_utility_builder_two = ThompsonSampling(num_features=100)
    ts_utility_function_one = ts_utility_builder_one.build_utility_function(
        posteriors=posteriors, datasets=datasets, key=key
    )
    ts_utility_function_two = ts_utility_builder_two.build_utility_function(
        posteriors=posteriors, datasets=datasets, key=key
    )
    test_key, _ = jr.split(key)
    test_X = test_target_function.generate_test_points(num_test_points, test_key)
    ts_utility_function_one_values = ts_utility_function_one(test_X)
    ts_utility_function_two_values = ts_utility_function_two(test_X)
    assert isinstance(ts_utility_function_one, Callable)
    assert isinstance(ts_utility_function_two, Callable)
    assert (ts_utility_function_one_values == ts_utility_function_two_values).all()


@pytest.mark.parametrize(
    "test_target_function",
    [(Forrester()), (LogarithmicGoldsteinPrice())],
)
@pytest.mark.parametrize("num_test_points", [50, 100])
@pytest.mark.parametrize("key", [jr.key(42), jr.key(10)])
@pytest.mark.filterwarnings(
    "ignore::UserWarning"
)  # Sampling with tfp causes JAX to raise a UserWarning due to some internal logic around jnp.argsort
def test_thompson_sampling_utility_function_different_key_different_function(
    test_target_function: AbstractContinuousTestFunction,
    num_test_points: int,
    key: KeyArray,
):
    dataset = test_target_function.generate_dataset(num_points=10, key=key)
    posterior = generate_dummy_conjugate_posterior(dataset)
    posteriors = {OBJECTIVE: posterior}
    datasets = {OBJECTIVE: dataset}
    sample_one_key = key
    sample_two_key, _ = jr.split(key)
    ts_utility_builder = ThompsonSampling(num_features=100)
    ts_utility_function_one = ts_utility_builder.build_utility_function(
        posteriors=posteriors, datasets=datasets, key=sample_one_key
    )
    ts_utility_function_two = ts_utility_builder.build_utility_function(
        posteriors=posteriors, datasets=datasets, key=sample_two_key
    )
    test_key, _ = jr.split(sample_two_key)
    test_X = test_target_function.generate_test_points(num_test_points, test_key)
    ts_utility_function_one_values = ts_utility_function_one(test_X)
    ts_utility_function_two_values = ts_utility_function_two(test_X)
    assert isinstance(ts_utility_function_one, Callable)
    assert isinstance(ts_utility_function_two, Callable)
    assert not (ts_utility_function_one_values == ts_utility_function_two_values).all()



# Copyright 2024 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from jax import config

config.update("jax_enable_x64", True)

from jax_decision_making.test_functions.continuous_functions import (
    AbstractContinuousTestFunction,
    Forrester,
    LogarithmicGoldsteinPrice,
)
from jax_decision_making.utility_functions.expected_improvement import (
    ExpectedImprovement,
)
from jax_decision_making.utils import (
    OBJECTIVE,
    get_best_latent_observation_val,
)
from gpjax.typing import KeyArray
import jax.numpy as jnp
import jax.random as jr
import pytest
import tensorflow_probability.substrates.jax as tfp

from tests.utils import generate_dummy_conjugate_posterior


@pytest.mark.parametrize(
    "test_target_function",
    [Forrester(), LogarithmicGoldsteinPrice()],
)
@pytest.mark.parametrize("key", [jr.PRNGKey(42), jr.PRNGKey(10)])
def test_expected_improvement_utility_function_correct_values(
    test_target_function: AbstractContinuousTestFunction,
    key: KeyArray,
):
    # Test validity of computed values with Monte-Carlo
    dataset = test_target_function.generate_dataset(num_points=10, key=key)
    posterior = generate_dummy_conjugate_posterior(dataset, test_target_function)
    posteriors = {OBJECTIVE: posterior}
    datasets = {OBJECTIVE: dataset}
    ei_fn = ExpectedImprovement().build_utility_function(posteriors, datasets, key)
    test_x = test_target_function.generate_test_points(100, key)
    ei = ei_fn(test_x)
    latent_dist = posterior.predict(test_x, dataset)
    pred_dist = posterior.likelihood(latent_dist)
    pred_mean = pred_dist.mean()
    pred_var = pred_dist.variance()
    samples = tfp.distributions.Normal(loc=pred_mean, scale=jnp.sqrt(pred_var)).sample(
        1000, seed=key
    )
    eta = get_best_latent_observation_val(posterior, dataset)
    mc_ei = jnp.expand_dims(jnp.mean(jnp.maximum(eta - samples, 0), 0), -1)
    assert jnp.all(ei >= 0)
    assert jnp.allclose(ei, mc_ei, rtol=0.08, atol=1e-6)
